from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import apply_baseline, load_baseline, resolve_baseline_path, write_baseline

    return apply_baseline, load_baseline, resolve_baseline_path, write_baseline


def test_baseline_write_and_apply(tmp_path: Path) -> None:
    apply_baseline, load_baseline, resolve_baseline_path, write_baseline = _load()
    baseline_path = resolve_baseline_path("baseline.txt", tmp_path)
    assert baseline_path == tmp_path / "baseline.txt"
    violations = ["violation-a", "violation-b"]
    write_baseline(baseline_path, violations)
    loaded = load_baseline(baseline_path)
    assert loaded == {"violation-a", "violation-b"}
    new, suppressed = apply_baseline(
        ["violation-a", "violation-c", "violation-b"], loaded
    )
    assert new == ["violation-c"]
    assert set(suppressed) == {"violation-a", "violation-b"}



from __future__ import annotations

from pathlib import Path

import libcst as cst

from gabion.refactor.model import FieldSpec, RefactorPlan, RefactorRequest, TextEdit


class RefactorEngine:
    def __init__(self, project_root: Path | None = None) -> None:
        self.project_root = project_root

    def plan_protocol_extraction(self, request: RefactorRequest) -> RefactorPlan:
        path = Path(request.target_path)
        if self.project_root and not path.is_absolute():
            path = self.project_root / path
        try:
            source = path.read_text()
        except Exception as exc:
            return RefactorPlan(errors=[f"Failed to read {path}: {exc}"])
        try:
            module = cst.parse_module(source)
        except Exception as exc:
            return RefactorPlan(errors=[f"LibCST parse failed for {path}: {exc}"])
        protocol = (request.protocol_name or "").strip()
        if not protocol:
            return RefactorPlan(errors=["Protocol name is required for extraction."])
        bundle = [name.strip() for name in request.bundle or [] if name.strip()]
        field_specs: list[FieldSpec] = []
        seen_fields: set[str] = set()
        for spec in request.fields or []:
            name = (spec.name or "").strip()
            if not name or name in seen_fields:
                continue
            seen_fields.add(name)
            field_specs.append(spec)
        if bundle:
            for name in bundle:
                if name in seen_fields:
                    continue
                seen_fields.add(name)
                field_specs.append(FieldSpec(name=name))
        elif field_specs:
            bundle = [spec.name for spec in field_specs]
        if not bundle:
            return RefactorPlan(errors=["Bundle fields are required for extraction."])

        body = list(module.body)

        insert_idx = _find_import_insert_index(body)

        protocol_base: cst.CSTNode
        import_stmt: cst.SimpleStatementLine | None = None
        if _has_typing_import(body):
            protocol_base = cst.Attribute(cst.Name("typing"), cst.Name("Protocol"))
        elif _has_typing_protocol_import(body):
            protocol_base = cst.Name("Protocol")
        else:
            protocol_base = cst.Name("Protocol")
            import_stmt = cst.SimpleStatementLine(
                [cst.ImportFrom(module=cst.Name("typing"), names=[cst.ImportAlias(name=cst.Name("Protocol"))])]
            )

        doc_lines = []
        if request.rationale:
            doc_lines.append(f"Rationale: {request.rationale}")
        doc_lines.append(f"Bundle: {', '.join(bundle)}")
        docstring = cst.SimpleStatementLine(
            [cst.Expr(cst.SimpleString('"""' + "\\n".join(doc_lines) + '"""'))]
        )
        warnings: list[str] = []

        def _annotation_for(hint: str | None) -> cst.BaseExpression:
            if not hint:
                return cst.Name("object")
            try:
                return cst.parse_expression(hint)
            except Exception as exc:
                warnings.append(f"Failed to parse type hint '{hint}': {exc}")
                return cst.Name("object")

        field_lines = []
        for spec in field_specs:
            field_lines.append(
                cst.SimpleStatementLine(
                    [
                        cst.AnnAssign(
                            target=cst.Name(spec.name),
                            annotation=cst.Annotation(_annotation_for(spec.type_hint)),
                            value=None,
                        )
                    ]
                )
            )
        class_body = [docstring] + field_lines
        class_def = cst.ClassDef(
            name=cst.Name(protocol),
            bases=[cst.Arg(protocol_base)],
            body=cst.IndentedBlock(body=class_body),
        )

        new_body = list(body)
        if import_stmt is not None:
            new_body.insert(insert_idx, import_stmt)
            insert_idx += 1
        if insert_idx > 0 and not isinstance(new_body[insert_idx - 1], cst.EmptyLine):
            new_body.insert(insert_idx, cst.EmptyLine())
            insert_idx += 1
        new_body.insert(insert_idx, class_def)

        new_module = module.with_changes(body=new_body)

        targets = {name.strip() for name in request.target_functions or [] if name.strip()}
        bundle_fields = [spec.name for spec in field_specs]
        protocol_hint = protocol
        if targets:
            compat_shim = bool(request.compatibility_shim)
            if compat_shim:
                new_module = _ensure_compat_imports(new_module)
            transformer = _RefactorTransformer(
                targets=targets,
                bundle_fields=bundle_fields,
                protocol_hint=protocol_hint,
                compat_shim=compat_shim,
            )
            new_module = new_module.visit(transformer)
            warnings.extend(transformer.warnings)
        if targets:
            target_module = _module_name(path, self.project_root)
            call_warnings, call_edits = _rewrite_call_sites(
                new_module,
                file_path=path,
                target_path=path,
                target_module=target_module,
                protocol_name=protocol,
                bundle_fields=bundle_fields,
                targets=targets,
            )
            warnings.extend(call_warnings)
            if call_edits is None:
                new_source = new_module.code
            else:
                new_module = call_edits
                new_source = new_module.code
        else:
            new_source = new_module.code
        if new_source == source:  # pragma: no cover
            warnings.append("No changes generated for protocol extraction.")  # pragma: no cover
            return RefactorPlan(warnings=warnings)  # pragma: no cover
        end_line = len(source.splitlines())
        edits = [
            TextEdit(
                path=str(path),
                start=(0, 0),
                end=(end_line, 0),
                replacement=new_source,
            )
        ]

        if targets and self.project_root:
            extra_edits, extra_warnings = _rewrite_call_sites_in_project(
                project_root=self.project_root,
                target_path=path,
                target_module=_module_name(path, self.project_root),
                protocol_name=protocol,
                bundle_fields=bundle_fields,
                targets=targets,
            )
            edits.extend(extra_edits)
            warnings.extend(extra_warnings)
        return RefactorPlan(edits=edits, warnings=warnings)


def _module_name(path: Path, project_root: Path | None) -> str:
    rel = path.with_suffix("")
    if project_root is not None:
        try:
            rel = rel.relative_to(project_root)
        except ValueError:
            pass
    parts = list(rel.parts)
    if parts and parts[0] == "src":
        parts = parts[1:]
    return ".".join(parts)


def _is_docstring(stmt: cst.CSTNode) -> bool:
    if not isinstance(stmt, cst.SimpleStatementLine) or not stmt.body:
        return False
    expr = stmt.body[0]
    return isinstance(expr, cst.Expr) and isinstance(expr.value, cst.SimpleString)


def _is_import(stmt: cst.CSTNode) -> bool:
    if not isinstance(stmt, cst.SimpleStatementLine):
        return False
    return any(isinstance(item, (cst.Import, cst.ImportFrom)) for item in stmt.body)


def _find_import_insert_index(body: list[cst.CSTNode]) -> int:
    insert_idx = 0
    if body and _is_docstring(body[0]):
        insert_idx = 1
    while insert_idx < len(body) and _is_import(body[insert_idx]):
        insert_idx += 1
    return insert_idx


def _module_expr_to_str(expr: cst.BaseExpression | None) -> str | None:
    if expr is None:
        return None
    if isinstance(expr, cst.Name):
        return expr.value
    if isinstance(expr, cst.Attribute):
        parts = []
        current: cst.BaseExpression | None = expr
        while isinstance(current, cst.Attribute):
            if isinstance(current.attr, cst.Name):
                parts.append(current.attr.value)
            current = current.value
        if isinstance(current, cst.Name):
            parts.append(current.value)
        if parts:
            return ".".join(reversed(parts))
    return None  # pragma: no cover


def _has_typing_import(body: list[cst.CSTNode]) -> bool:
    for stmt in body:
        if not isinstance(stmt, cst.SimpleStatementLine):
            continue
        for item in stmt.body:
            if isinstance(item, cst.Import):
                for alias in item.names:
                    if isinstance(alias, cst.ImportAlias) and isinstance(alias.name, cst.Name):
                        if alias.name.value == "typing":
                            return True
                    if isinstance(alias, cst.ImportAlias) and isinstance(
                        alias.name, cst.Attribute
                    ):
                        if _module_expr_to_str(alias.name) == "typing":
                            return True  # pragma: no cover
    return False


def _has_typing_protocol_import(body: list[cst.CSTNode]) -> bool:
    for stmt in body:
        if not isinstance(stmt, cst.SimpleStatementLine):
            continue
        for item in stmt.body:
            if not isinstance(item, cst.ImportFrom):
                continue
            module = _module_expr_to_str(item.module)
            if module != "typing":
                continue
            for alias in item.names:
                if isinstance(alias, cst.ImportAlias) and isinstance(alias.name, cst.Name):
                    if alias.name.value == "Protocol":
                        return True
    return False


def _has_typing_overload_import(body: list[cst.CSTNode]) -> bool:
    for stmt in body:
        if not isinstance(stmt, cst.SimpleStatementLine):
            continue
        for item in stmt.body:
            if not isinstance(item, cst.ImportFrom):
                continue
            module = _module_expr_to_str(item.module)
            if module != "typing":
                continue
            for alias in item.names:
                if isinstance(alias, cst.ImportAlias) and isinstance(alias.name, cst.Name):
                    if alias.name.value == "overload":
                        return True
    return False


def _has_warnings_import(body: list[cst.CSTNode]) -> bool:
    for stmt in body:
        if not isinstance(stmt, cst.SimpleStatementLine):
            continue
        for item in stmt.body:
            if isinstance(item, cst.Import):
                for alias in item.names:
                    if isinstance(alias, cst.ImportAlias) and isinstance(alias.name, cst.Name):
                        if alias.name.value == "warnings":
                            return True
    return False


def _ensure_compat_imports(module: cst.Module) -> cst.Module:
    body = list(module.body)
    insert_idx = _find_import_insert_index(body)
    if not _has_warnings_import(body):
        body.insert(
            insert_idx,
            cst.SimpleStatementLine(
                [cst.Import(names=[cst.ImportAlias(name=cst.Name("warnings"))])]
            ),
        )
        insert_idx += 1
    if not _has_typing_overload_import(body):
        body.insert(
            insert_idx,
            cst.SimpleStatementLine(
                [
                    cst.ImportFrom(
                        module=cst.Name("typing"),
                        names=[cst.ImportAlias(name=cst.Name("overload"))],
                    )
                ]
            ),
        )
        insert_idx += 1
    return module.with_changes(body=body)


def _collect_import_context(
    module: cst.Module,
    *,
    target_module: str,
    protocol_name: str,
) -> tuple[dict[str, str], dict[str, str], str | None]:
    module_aliases: dict[str, str] = {}
    imported_targets: dict[str, str] = {}
    protocol_alias: str | None = None
    for stmt in module.body:
        if not isinstance(stmt, cst.SimpleStatementLine):
            continue
        for item in stmt.body:
            if isinstance(item, cst.Import):
                for alias in item.names:
                    if not isinstance(alias, cst.ImportAlias):  # pragma: no cover
                        continue
                    module_name = _module_expr_to_str(alias.name)
                    if not module_name:  # pragma: no cover
                        continue
                    if module_name != target_module:
                        continue
                    local = alias.asname.name.value if alias.asname else module_name
                    module_aliases[local] = module_name
            elif isinstance(item, cst.ImportFrom):
                module_name = _module_expr_to_str(item.module)
                if module_name != target_module:
                    continue
                for alias in item.names:
                    if not isinstance(alias, cst.ImportAlias):
                        continue
                    if not isinstance(alias.name, cst.Name):
                        continue
                    local = alias.asname.name.value if alias.asname else alias.name.value
                    imported_targets[local] = alias.name.value
                    if alias.name.value == protocol_name:
                        protocol_alias = local
    return module_aliases, imported_targets, protocol_alias


def _rewrite_call_sites(
    module: cst.Module,
    *,
    file_path: Path,
    target_path: Path,
    target_module: str,
    protocol_name: str,
    bundle_fields: list[str],
    targets: set[str],
) -> tuple[list[str], cst.Module | None]:
    warnings: list[str] = []
    file_is_target = file_path == target_path
    if not targets:
        return warnings, None
    target_simple = {name for name in targets if "." not in name}
    target_methods: dict[str, set[str]] = {}
    for name in targets:
        if "." not in name:
            continue
        parts = name.split(".")
        class_name = ".".join(parts[:-1])
        method = parts[-1]
        target_methods.setdefault(class_name, set()).add(method)
    module_aliases: dict[str, str] = {}
    imported_targets: dict[str, str] = {}
    protocol_alias: str | None = None
    if not file_is_target and target_module:
        module_aliases, imported_targets, protocol_alias = _collect_import_context(
            module, target_module=target_module, protocol_name=protocol_name
        )

    constructor_expr: cst.BaseExpression
    needs_import = False
    if file_is_target:
        constructor_expr = cst.Name(protocol_name)
    else:
        if protocol_alias:
            constructor_expr = cst.Name(protocol_alias)
        elif module_aliases:
            alias = sorted(module_aliases.keys())[0]
            constructor_expr = cst.Attribute(cst.Name(alias), cst.Name(protocol_name))
        else:
            constructor_expr = cst.Name(protocol_name)
            needs_import = True

    transformer = _CallSiteTransformer(
        file_is_target=file_is_target,
        target_simple=target_simple,
        target_methods=target_methods,
        module_aliases=set(module_aliases.keys()),
        imported_targets=set(
            name for name, original in imported_targets.items() if original in target_simple
        ),
        bundle_fields=bundle_fields,
        constructor_expr=constructor_expr,
    )
    new_module = module.visit(transformer)
    warnings.extend(transformer.warnings)
    if not transformer.changed:
        return warnings, None

    if not file_is_target and needs_import and target_module:
        body = list(new_module.body)
        insert_idx = _find_import_insert_index(body)
        try:
            module_expr = cst.parse_expression(target_module)
        except Exception:  # pragma: no cover
            module_expr = cst.Name(target_module.split(".")[0])  # pragma: no cover
        if not isinstance(module_expr, (cst.Name, cst.Attribute)):
            module_expr = cst.Name(target_module.split(".")[0])  # pragma: no cover
        import_stmt = cst.SimpleStatementLine(
            [
                cst.ImportFrom(
                    module=module_expr,
                    names=[cst.ImportAlias(name=cst.Name(protocol_name))],
                )
            ]
        )
        body.insert(insert_idx, import_stmt)
        new_module = new_module.with_changes(body=body)
    return warnings, new_module


def _rewrite_call_sites_in_project(
    *,
    project_root: Path,
    target_path: Path,
    target_module: str,
    protocol_name: str,
    bundle_fields: list[str],
    targets: set[str],
) -> tuple[list[TextEdit], list[str]]:
    edits: list[TextEdit] = []
    warnings: list[str] = []
    scan_root = project_root / "src"
    if not scan_root.exists():
        scan_root = project_root
    for path in sorted(scan_root.rglob("*.py")):
        if path == target_path:
            continue
        try:
            source = path.read_text()
        except Exception as exc:
            warnings.append(f"Failed to read {path}: {exc}")
            continue
        try:
            module = cst.parse_module(source)
        except Exception as exc:
            warnings.append(f"LibCST parse failed for {path}: {exc}")
            continue
        call_warnings, updated_module = _rewrite_call_sites(
            module,
            file_path=path,
            target_path=target_path,
            target_module=target_module,
            protocol_name=protocol_name,
            bundle_fields=bundle_fields,
            targets=targets,
        )
        warnings.extend(call_warnings)
        if updated_module is None:
            continue
        new_source = updated_module.code
        if new_source == source:  # pragma: no cover
            continue
        end_line = len(source.splitlines())
        edits.append(
            TextEdit(
                path=str(path),
                start=(0, 0),
                end=(end_line, 0),
                replacement=new_source,
            )
        )
    return edits, warnings


class _RefactorTransformer(cst.CSTTransformer):
    def __init__(
        self,
        *,
        targets: set[str],
        bundle_fields: list[str],
        protocol_hint: str,
        compat_shim: bool = False,
    ) -> None:
        # dataflow-bundle: bundle_fields, compat_shim, protocol_hint, targets
        self.targets = targets
        self.bundle_fields = bundle_fields
        self.protocol_hint = protocol_hint
        self.compat_shim = compat_shim
        self.warnings: list[str] = []
        self._stack: list[str] = []

    def visit_ClassDef(self, node: cst.ClassDef) -> bool:
        self._stack.append(node.name.value)
        return True

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.CSTNode:
        if self._stack:
            self._stack.pop()
        return updated_node

    def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
        self._stack.append(node.name.value)
        return True

    def leave_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> cst.CSTNode:
        updated = self._maybe_rewrite_function(original_node, updated_node)
        if self._stack:
            self._stack.pop()
        return updated

    def visit_AsyncFunctionDef(self, node: cst.AsyncFunctionDef) -> bool:  # pragma: no cover
        self._stack.append(node.name.value)  # pragma: no cover
        return True  # pragma: no cover

    def leave_AsyncFunctionDef(  # pragma: no cover
        self, original_node: cst.AsyncFunctionDef, updated_node: cst.AsyncFunctionDef
    ) -> cst.CSTNode:
        updated = self._maybe_rewrite_function(original_node, updated_node)  # pragma: no cover
        if self._stack:  # pragma: no cover
            self._stack.pop()  # pragma: no cover
        return updated  # pragma: no cover

    def _maybe_rewrite_function(
        self,
        original_node: cst.FunctionDef,
        updated_node: cst.FunctionDef | cst.AsyncFunctionDef,
    ) -> cst.CSTNode:
        qualname = ".".join(self._stack)
        name = original_node.name.value
        if name not in self.targets and qualname not in self.targets:
            return updated_node

        original_params = self._ordered_param_names(original_node.params)
        if not original_params:
            return updated_node

        keep_self = original_params[0] in {"self", "cls"}
        self_param = None
        if keep_self:
            self_param = self._find_self_param(original_node.params, original_params[0])

        bundle_param_name = self._choose_bundle_name(original_params)
        new_params = self._build_parameters(self_param, bundle_param_name)

        bundle_set = set(self.bundle_fields)
        target_fields = [
            name
            for name in original_params
            if name in bundle_set and name not in {"self", "cls"}
        ]
        new_body = self._inject_preamble(
            updated_node.body, bundle_param_name, target_fields
        )
        updated_impl = updated_node.with_changes(params=new_params, body=new_body)
        if not self.compat_shim:
            return updated_impl
        impl_name = f"_{name}_bundle"
        impl_node = updated_impl.with_changes(
            name=cst.Name(impl_name),
            decorators=[],
        )
        shim_nodes = self._build_compat_shim(
            original_node=original_node,
            impl_name=impl_name,
            bundle_param=bundle_param_name,
            self_param=self_param,
        )
        return cst.FlattenSentinel([*shim_nodes, impl_node])

    def _build_compat_shim(
        self,
        *,
        original_node: cst.FunctionDef | cst.AsyncFunctionDef,
        impl_name: str,
        bundle_param: str,
        self_param: cst.Param | None,
    ) -> list[cst.CSTNode]:
        decorators = list(original_node.decorators)
        overload_decorators = [cst.Decorator(cst.Name("overload")), *decorators]
        bundle_params = self._build_parameters(self_param, bundle_param)
        legacy_params = original_node.params
        bundle_stub = self._build_overload_stub(
            original_node=original_node,
            params=bundle_params,
            decorators=overload_decorators,
        )
        legacy_stub = self._build_overload_stub(
            original_node=original_node,
            params=legacy_params,
            decorators=overload_decorators,
        )
        wrapper_params = self._build_shim_parameters(self_param)
        wrapper_body = self._build_shim_body(
            impl_name=impl_name,
            bundle_type=self.protocol_hint,
            self_param=self_param,
            is_async=bool(original_node.asynchronous),
            public_name=original_node.name.value,
        )
        wrapper = self._build_wrapper(
            original_node=original_node,
            params=wrapper_params,
            body=wrapper_body,
            decorators=decorators,
        )
        return [bundle_stub, legacy_stub, wrapper]

    def _build_overload_stub(
        self,
        *,
        original_node: cst.FunctionDef,
        params: cst.Parameters,
        decorators: list[cst.Decorator],
    ) -> cst.CSTNode:
        body = cst.IndentedBlock(
            body=[cst.SimpleStatementLine([cst.Expr(cst.Ellipsis())])]
        )
        return original_node.with_changes(
            params=params,
            body=body,
            decorators=decorators,
        )

    def _build_wrapper(
        self,
        *,
        original_node: cst.FunctionDef,
        params: cst.Parameters,
        body: cst.BaseSuite,
        decorators: list[cst.Decorator],
    ) -> cst.CSTNode:
        return original_node.with_changes(
            params=params,
            body=body,
            decorators=decorators,
        )

    def _build_shim_parameters(self, self_param: cst.Param | None) -> cst.Parameters:
        params: list[cst.Param] = []
        if self_param is not None:
            params.append(self_param)
        return cst.Parameters(
            params=params,
            star_arg=cst.Param(name=cst.Name("args")),
            kwonly_params=[],
            star_kwarg=cst.Param(name=cst.Name("kwargs")),
            posonly_params=[],
            posonly_ind=cst.MaybeSentinel.DEFAULT,
        )

    def _build_shim_body(
        self,
        *,
        impl_name: str,
        bundle_type: str,
        self_param: cst.Param | None,
        is_async: bool,
        public_name: str,
    ) -> cst.BaseSuite:
        receiver = self_param.name.value if self_param is not None else None
        impl_call = f"{receiver}.{impl_name}" if receiver else impl_name
        return_prefix = "return await" if is_async else "return"
        guard = (
            f"if args and isinstance(args[0], {bundle_type}):\n"
            f"    if len(args) != 1 or kwargs:\n"
            f"        raise TypeError(\"{public_name}() bundle call expects a single {bundle_type} argument\")\n"
            f"    {return_prefix} {impl_call}(args[0])\n"
        )
        warn = (
            f"warnings.warn(\"{public_name}() is deprecated; use {public_name}({bundle_type}(...))\", "
            "DeprecationWarning, stacklevel=2)"
        )
        build = f"bundle = {bundle_type}(*args, **kwargs)"
        tail = f"{return_prefix} {impl_call}(bundle)"
        return cst.IndentedBlock(
            body=[
                cst.parse_statement(guard),
                cst.parse_statement(warn),
                cst.parse_statement(build),
                cst.parse_statement(tail),
            ]
        )

    def _ordered_param_names(self, params: cst.Parameters) -> list[str]:
        names: list[str] = []
        for param in params.posonly_params:
            names.append(param.name.value)
        for param in params.params:
            names.append(param.name.value)
        for param in params.kwonly_params:
            names.append(param.name.value)
        return names

    def _find_self_param(
        self, params: cst.Parameters, name: str
    ) -> cst.Param | None:
        for param in params.posonly_params:
            if param.name.value == name:
                return param
        for param in params.params:
            if param.name.value == name:
                return param
        return None

    def _choose_bundle_name(self, existing: list[str]) -> str:
        candidate = "bundle"
        if candidate not in existing:
            return candidate
        idx = 1
        while f"bundle_{idx}" in existing:
            idx += 1
        return f"bundle_{idx}"

    def _build_parameters(
        self, self_param: cst.Param | None, bundle_name: str
    ) -> cst.Parameters:
        params: list[cst.Param] = []
        if self_param is not None:
            params.append(self_param)
        annotation = None
        if self.protocol_hint:
            try:
                annotation = cst.Annotation(cst.parse_expression(self.protocol_hint))
            except Exception as exc:
                self.warnings.append(
                    f"Failed to parse protocol type hint '{self.protocol_hint}': {exc}"
                )
        params.append(
            cst.Param(
                name=cst.Name(bundle_name),
                annotation=annotation,
            )
        )
        return cst.Parameters(
            params=params,
            star_arg=cst.MaybeSentinel.DEFAULT,
            kwonly_params=[],
            star_kwarg=None,
            posonly_params=[],
            posonly_ind=cst.MaybeSentinel.DEFAULT,
        )

    def _inject_preamble(
        self, body: cst.BaseSuite, bundle_name: str, fields: list[str]
    ) -> cst.BaseSuite:
        if not fields:
            return body
        if not isinstance(body, cst.IndentedBlock):
            return body
        assign_lines = [
            cst.SimpleStatementLine(
                [
                    cst.Assign(
                        targets=[cst.AssignTarget(cst.Name(name))],
                        value=cst.Attribute(
                            value=cst.Name(bundle_name),
                            attr=cst.Name(name),
                        ),
                    )
                ]
            )
            for name in fields
        ]
        existing = list(body.body)
        insert_at = 0
        if existing:
            first = existing[0]
            if isinstance(first, cst.SimpleStatementLine) and first.body:
                expr = first.body[0]
                if isinstance(expr, cst.Expr) and isinstance(expr.value, cst.SimpleString):
                    insert_at = 1
        new_body = existing[:insert_at] + assign_lines + existing[insert_at:]
        return body.with_changes(body=new_body)


class _CallSiteTransformer(cst.CSTTransformer):
    def __init__(
        self,
        *,
        file_is_target: bool,
        target_simple: set[str],
        target_methods: dict[str, set[str]],
        module_aliases: set[str],
        imported_targets: set[str],
        bundle_fields: list[str],
        constructor_expr: cst.BaseExpression,
    ) -> None:
        # dataflow-bundle: bundle_fields, constructor_expr, file_is_target, imported_targets, module_aliases, target_methods, target_simple
        self.file_is_target = file_is_target
        self.target_simple = target_simple
        self.target_methods = target_methods
        self.module_aliases = module_aliases
        self.imported_targets = imported_targets
        self.bundle_fields = bundle_fields
        self.constructor_expr = constructor_expr
        self.changed = False
        self.warnings: list[str] = []
        self._class_stack: list[str] = []

    def visit_ClassDef(self, node: cst.ClassDef) -> bool:
        self._class_stack.append(node.name.value)
        return True

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.CSTNode:
        if self._class_stack:
            self._class_stack.pop()
        return updated_node

    def leave_Call(self, original_node: cst.Call, updated_node: cst.Call) -> cst.CSTNode:
        if not self._is_target_call(updated_node.func):
            return updated_node
        if self._already_wrapped(updated_node):
            return updated_node
        bundle_args = self._build_bundle_args(updated_node)
        if bundle_args is None:
            return updated_node
        bundle_call = cst.Call(func=self.constructor_expr, args=bundle_args)
        self.changed = True
        return updated_node.with_changes(args=[cst.Arg(value=bundle_call)])

    def _is_target_call(self, func: cst.BaseExpression) -> bool:
        if isinstance(func, cst.Name):
            if self.file_is_target and func.value in self.target_simple:
                return True
            if not self.file_is_target and func.value in self.imported_targets:
                return True
            return False
        if isinstance(func, cst.Attribute):
            if not isinstance(func.attr, cst.Name):  # pragma: no cover
                return False
            attr = func.attr.value
            if self.file_is_target and self._class_stack:
                class_name = ".".join(self._class_stack)
                methods = self.target_methods.get(class_name, set())
                if attr in methods and isinstance(func.value, cst.Name):
                    if func.value.value in {"self", "cls", self._class_stack[-1]}:
                        return True
            if not self.file_is_target and isinstance(func.value, cst.Name):
                if func.value.value in self.module_aliases and attr in self.target_simple:
                    return True
        return False

    def _already_wrapped(self, call: cst.Call) -> bool:
        if len(call.args) != 1:
            return False
        arg = call.args[0]
        if arg.star:
            return False
        value = arg.value
        if not isinstance(value, cst.Call):
            return False
        if isinstance(value.func, cst.Name) and isinstance(self.constructor_expr, cst.Name):
            return value.func.value == self.constructor_expr.value
        if isinstance(value.func, cst.Attribute) and isinstance(self.constructor_expr, cst.Attribute):
            return value.func.attr.value == self.constructor_expr.attr.value
        return False

    def _build_bundle_args(self, call: cst.Call) -> list[cst.Arg] | None:
        if any(arg.star in {"*", "**"} for arg in call.args):
            self.warnings.append("Skipped call with star args/kwargs during refactor.")
            return None
        positional = [arg for arg in call.args if arg.keyword is None]
        keyword_args = {
            arg.keyword.value: arg.value
            for arg in call.args
            if arg.keyword is not None and isinstance(arg.keyword, cst.Name)
        }
        for key in keyword_args:
            if key not in self.bundle_fields:
                self.warnings.append(
                    f"Skipped call with unknown keyword '{key}' during refactor."
                )
                return None
        mapping: dict[str, cst.BaseExpression] = {}
        mapping.update(keyword_args)
        remaining = [field for field in self.bundle_fields if field not in mapping]
        if len(positional) > len(remaining):
            self.warnings.append("Skipped call with extra positional args during refactor.")
            return None
        for field, arg in zip(remaining, positional):
            mapping[field] = arg.value
        if len(mapping) != len(self.bundle_fields):
            self.warnings.append("Skipped call with missing bundle fields during refactor.")
            return None
        return [
            cst.Arg(keyword=cst.Name(field), value=mapping[field])
            for field in self.bundle_fields
        ]



from __future__ import annotations

import ast
from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def test_collect_module_exports_augassign_only() -> None:
    da = _load()
    module = ast.parse("__all__ += ['A']\nA = 1\n")
    export_names, export_map = da._collect_module_exports(
        module,
        module_name="demo",
        import_map={},
    )
    assert export_names == {"A"}
    assert export_map["A"] == "demo.A"


def test_base_identifier_invalid_attribute() -> None:
    da = _load()
    bad_attr = ast.Attribute(
        value=ast.Constant(value=1),
        attr=None,
        ctx=ast.Load(),
    )
    assert da._base_identifier(bad_attr) is None


def test_collect_return_aliases_empty_params() -> None:
    da = _load()
    tree = ast.parse(
        "def f():\n"
        "    return 1\n"
    )
    fn = tree.body[0]
    parents = {}
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            parents[child] = node
    aliases = da._collect_return_aliases([fn], parents, ignore_params=None)
    assert aliases == {}


def test_collect_return_aliases_all_params_ignored() -> None:
    da = _load()
    tree = ast.parse(
        "def f(a):\n"
        "    return a\n"
    )
    fn = tree.body[0]
    parents = {}
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            parents[child] = node
    aliases = da._collect_return_aliases([fn], parents, ignore_params={"a"})
    assert aliases == {}


def test_analyze_file_recursive_false(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text(
        "def callee(x):\n"
        "    return x\n"
        "\n"
        "def caller(a, b):\n"
        "    callee(a)\n"
        "    callee(b)\n"
    )
    groups, spans = da.analyze_file(target, recursive=False)
    assert groups
    assert spans


def test_resolve_local_callee_ambiguity_and_scope(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text(
        textwrap.dedent(
            """
            def outer():
                def inner(x):
                    return x
                def inner(x):
                    return x
                def caller(y):
                    return inner(y)
                return caller

            def helper(x):
                return x

            def global_caller(z):
                return helper(z)

            def outer2():
                def nested(x):
                    return x
                return nested

            def top():
                nested(1)
            """
        ).strip()
        + "\n"
    )
    config = da.AuditConfig(project_root=tmp_path)
    groups, _ = da.analyze_file(target, recursive=True, config=config)
    assert groups


def test_analyze_file_local_callee_globals(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text(
        textwrap.dedent(
            """
            def helper(x):
                return x

            def outer():
                def inner(y):
                    return helper(y)
                return inner

            def caller(z):
                return helper(z)
            """
        ).strip()
        + "\n"
    )
    groups, _ = da.analyze_file(target, recursive=True, config=da.AuditConfig(project_root=tmp_path))
    assert groups


def test_resolve_callee_imports_and_self(tmp_path: Path) -> None:
    da = _load()
    pkg_root = tmp_path / "pkg"
    pkg_root.mkdir()
    mod_a = pkg_root / "a.py"
    mod_b = pkg_root / "b.py"
    mod_a.write_text(
        textwrap.dedent(
            """
            from pkg.b import ext

            class C:
                def method(self, x):
                    return x

                def call(self, y):
                    return self.method(y)

            def caller(z):
                return ext(z)
            """
        ).strip()
        + "\n"
    )
    mod_b.write_text(
        "def ext(v):\n"
        "    return v\n"
    )
    paths = [mod_a, mod_b]
    by_name, by_qual = da._build_function_index(
        paths,
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        transparent_decorators=None,
    )
    symbol_table = da._build_symbol_table(paths, tmp_path, external_filter=True)
    class_index = da._collect_class_index(paths, tmp_path)
    caller_info = by_name["caller"][0]
    resolved = da._resolve_callee(
        "ext",
        caller_info,
        by_name,
        by_qual,
        symbol_table,
        tmp_path,
        class_index,
    )
    assert resolved is not None
    method_info = by_name["call"][0]
    resolved_method = da._resolve_callee(
        "self.method",
        method_info,
        by_name,
        by_qual,
        symbol_table,
        tmp_path,
        class_index,
    )
    assert resolved_method is not None


def test_resolve_callee_external_filtered(tmp_path: Path) -> None:
    da = _load()
    mod_a = tmp_path / "a.py"
    mod_a.write_text(
        "import external.mod as ext\n"
        "def caller(x):\n"
        "    return ext.run(x)\n"
    )
    paths = [mod_a]
    by_name, by_qual = da._build_function_index(
        paths,
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        transparent_decorators=None,
    )
    symbol_table = da._build_symbol_table(paths, tmp_path, external_filter=True)
    class_index = da._collect_class_index(paths, tmp_path)
    caller_info = by_name["caller"][0]
    resolved = da._resolve_callee(
        "ext.run",
        caller_info,
        by_name,
        by_qual,
        symbol_table,
        tmp_path,
        class_index,
    )
    assert resolved is None


def test_resolve_callee_global_and_imported(tmp_path: Path) -> None:
    da = _load()
    mod_path = tmp_path / "mod.py"
    caller = da.FunctionInfo(
        name="caller",
        qual="mod.caller",
        path=mod_path,
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
    )
    helper = da.FunctionInfo(
        name="helper",
        qual="mod.helper",
        path=mod_path,
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
    )
    by_name = {"helper": [helper], "caller": [caller]}
    by_qual = {helper.qual: helper, caller.qual: caller}
    resolved = da._resolve_callee(
        "helper",
        caller,
        by_name,
        by_qual,
        None,
        tmp_path,
        None,
    )
    assert resolved == helper

    imported = da.FunctionInfo(
        name="target",
        qual="other.target",
        path=tmp_path / "other.py",
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
    )
    by_qual[imported.qual] = imported
    by_name.setdefault("target", []).append(imported)
    table = da.SymbolTable(external_filter=True)
    table.imports[("mod", "alias")] = imported.qual
    table.internal_roots.add("other")
    resolved = da._resolve_callee(
        "alias",
        caller,
        by_name,
        by_qual,
        table,
        tmp_path,
        None,
    )
    assert resolved == imported

    by_qual["mod.service.run"] = imported
    resolved = da._resolve_callee(
        "service.run",
        caller,
        by_name,
        by_qual,
        table,
        tmp_path,
        None,
    )
    assert resolved == imported

    resolved = da._resolve_callee(
        "mod.helper",
        caller,
        by_name,
        by_qual,
        None,
        tmp_path,
        None,
    )
    assert resolved == helper


def test_resolve_method_in_hierarchy_edges() -> None:
    da = _load()
    class_index = {
        "pkg.Base": da.ClassInfo(
            qual="pkg.Base", module="pkg", bases=["Missing"], methods={"run"}
        )
    }
    by_qual = {}
    assert (
        da._resolve_method_in_hierarchy(
            "pkg.Base",
            "run",
            class_index=class_index,
            by_qual=by_qual,
            symbol_table=None,
            seen={"pkg.Base"},
        )
        is None
    )
    assert (
        da._resolve_method_in_hierarchy(
            "pkg.Missing",
            "run",
            class_index=class_index,
            by_qual=by_qual,
            symbol_table=None,
            seen=set(),
        )
        is None
    )
    assert (
        da._resolve_method_in_hierarchy(
            "pkg.Base",
            "missing",
            class_index=class_index,
            by_qual=by_qual,
            symbol_table=None,
            seen=set(),
        )
        is None
    )


def test_render_helpers_and_baseline(tmp_path: Path) -> None:
    da = _load()
    assert da._infer_root({}) == Path(".")
    assert da._callee_key("") == ""
    assert (
        da._resolve_callee(
            "",
            da.FunctionInfo(
                name="caller",
                qual="caller",
                path=tmp_path / "mod.py",
                params=[],
                annots={},
                calls=[],
                unused_params=set(),
            ),
            {},
            {},
            None,
            tmp_path,
            None,
        )
        is None
    )
    assert da._resolve_class_candidates(
        "",
        module="",
        symbol_table=None,
        class_index={},
    ) == []

    base_dir = tmp_path / "baseline_dir"
    base_dir.mkdir()
    assert da._load_baseline(base_dir) == set()
    assert da._apply_baseline(["a"], set()) == (["a"], [])

    report = da.render_synthesis_section(
        {
            "protocols": [
                {
                    "name": "Bundle",
                    "tier": 2,
                    "fields": [{"name": "a", "type_hint": "Optional[int]"}],
                }
            ],
            "warnings": ["warn"],
            "errors": ["err"],
        }
    )
    assert "Synthesis plan" in report

    stubs = da.render_protocol_stubs(
        {
            "protocols": [
                {
                    "name": "Bundle",
                    "tier": 3,
                    "bundle": ["a", "b"],
                    "rationale": "test",
                    "fields": [
                        {"name": "a", "type_hint": "Optional[int]"},
                        {"name": "b", "type_hint": "Union[int, str]"},
                    ],
                },
                {
                    "name": "Empty",
                    "tier": 2,
                    "bundle": [],
                    "fields": [],
                },
            ],
            "warnings": [],
            "errors": [],
        },
        kind="unknown",
    )
    assert "TODO_Name_Me" in stubs

    plan = da.build_refactor_plan({}, [], config=da.AuditConfig())
    refactor_text = da.render_refactor_plan(plan)
    assert "No refactoring plan" in refactor_text

    plan = {
        "bundles": [{"bundle": ["a", "b"], "order": ["x"], "cycles": [["x"]]}],
        "warnings": ["warn"],
    }
    refactor_text = da.render_refactor_plan(plan)
    assert "Cycles:" in refactor_text
    assert "Warnings:" in refactor_text


def test_emit_report_tier2_violation(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "a.py"
    target.write_text("def caller(a, b):\n    return a\n")
    groups_by_path = {target: {"caller": [{"a", "b"}, {"a", "b"}]}}
    report, violations = da._emit_report(groups_by_path, max_components=10)
    assert "tier-2" in report
    assert violations


def test_infer_root_with_groups(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "a.py"
    groups_by_path = {target: {}}
    assert da._infer_root(groups_by_path) == target


def test_build_refactor_plan_skips_unresolved_and_opaque(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text(
        textwrap.dedent(
            """
            def callee(x):
                return x

            def opaque(fn):
                return fn

            @opaque
            def hidden(a, b):
                return a

            def caller(a, b):
                callee(a)
                callee(b)
                unknown(a)
                hidden(a, b)
            """
        ).strip()
        + "\n"
    )
    config = da.AuditConfig(
        project_root=tmp_path,
        exclude_dirs=set(),
        ignore_params=set(),
        external_filter=True,
        strictness="high",
        transparent_decorators={"transparent"},
    )
    groups, _ = da.analyze_file(target, recursive=True, config=config)
    groups_by_path = {target: groups}
    plan = da.build_refactor_plan(groups_by_path, [target], config=config)
    assert "bundles" in plan


def test_build_refactor_plan_skips_none_and_nontransparent(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text(
        textwrap.dedent(
            """
            def opaque(fn):
                return fn

            @opaque
            def hidden(a, b):
                return a

            def caller(a, b):
                hidden(a, b)
                unknown(a)
            """
        ).strip()
        + "\n"
    )
    config = da.AuditConfig(
        project_root=tmp_path,
        exclude_dirs=set(),
        ignore_params=set(),
        external_filter=True,
        strictness="high",
        transparent_decorators=None,
    )
    groups_by_path = {target: {"caller": [set(["a", "b"])]}}
    plan = da.build_refactor_plan(groups_by_path, [target], config=config)
    assert "bundles" in plan


def test_render_type_mermaid_edges() -> None:
    da = _load()
    graph = da._render_type_mermaid(
        ["bad entry"],
        ["missing format", "file:fn.param downstream types conflict: []"],
    )
    assert "flowchart LR" in graph


def test_compute_knob_param_names_non_const_kw(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text(
        "def callee(a, b):\n"
        "    return b\n"
        "\n"
        "def caller(x):\n"
        "    return callee(a=1, b=x + 1)\n"
    )
    paths = [target]
    by_name, by_qual = da._build_function_index(
        paths,
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        transparent_decorators=None,
    )
    symbol_table = da._build_symbol_table(paths, tmp_path, external_filter=True)
    class_index = da._collect_class_index(paths, tmp_path)
    knob_names = da._compute_knob_param_names(
        by_name=by_name,
        by_qual=by_qual,
        symbol_table=symbol_table,
        project_root=tmp_path,
        class_index=class_index,
        strictness="high",
    )
    assert "a" in knob_names or "b" in knob_names


def test_iter_dataclass_call_bundles(tmp_path: Path) -> None:
    da = _load()
    root = tmp_path / "src"
    root.mkdir()
    mod_a = root / "a.py"
    mod_b = root / "b.py"
    mod_a.write_text(
        textwrap.dedent(
            """
            from dataclasses import dataclass

            @dataclass
            class Bundle:
                a: int
                b: int
            """
        ).strip()
        + "\n"
    )
    mod_b.write_text(
        textwrap.dedent(
            """
            from a import Bundle

            def build():
                Bundle(1, 2, 3)
                Bundle(a=1, b=2)
                mod.Bundle(1, 2)
            """
        ).strip()
        + "\n"
    )
    paths = [mod_a, mod_b]
    symbol_table = da._build_symbol_table(paths, root, external_filter=True)
    registry = da._collect_dataclass_registry(paths, project_root=root)
    bundles = da._iter_dataclass_call_bundles(
        mod_b,
        project_root=root,
        symbol_table=symbol_table,
        dataclass_registry=registry,
    )
    assert tuple(sorted(("a", "b"))) in bundles
    bundles = da._iter_dataclass_call_bundles(
        mod_b,
        project_root=root,
        symbol_table=None,
        dataclass_registry={"b.Bundle": ["a", "b"]},
    )
    assert tuple(sorted(("a", "b"))) in bundles
    bundles = da._iter_dataclass_call_bundles(
        mod_b,
        project_root=root,
        symbol_table=None,
        dataclass_registry={"Bundle": ["a", "b"]},
    )
    assert tuple(sorted(("a", "b"))) in bundles
    table = da.SymbolTable(external_filter=True)
    table.star_imports["b"] = {"pkg"}
    table.module_exports["pkg"] = {"mod"}
    table.module_export_map["pkg"] = {"mod": "pkg.mod"}
    table.internal_roots.add("pkg")
    bundles = da._iter_dataclass_call_bundles(
        mod_b,
        project_root=root,
        symbol_table=table,
        dataclass_registry={"pkg.mod.Bundle": ["a", "b"]},
    )
    assert tuple(sorted(("a", "b"))) in bundles


def test_build_synthesis_plan_skips_empty_members(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text("def f(a):\n    return a\n")
    groups_by_path = {target: {"f": [set()]}}
    plan = da.build_synthesis_plan(
        groups_by_path,
        project_root=tmp_path,
        max_tier=2,
        min_bundle_size=1,
        allow_singletons=True,
        config=da.AuditConfig(project_root=tmp_path),
    )
    assert "protocols" in plan


def test_analyze_file_local_resolution_ambiguous(tmp_path: Path) -> None:
    da = _load()
    path = tmp_path / "mod.py"
    path.write_text(
        textwrap.dedent(
            """
            def outer(x):
                class C:
                    def foo(self, y):
                        return y

                class D:
                    def foo(self, y):
                        return y

                def caller(z):
                    return foo(z)

                return caller(x)
            """
        ).strip()
        + "\n"
    )
    groups, _ = da.analyze_file(
        path,
        recursive=True,
        config=da.AuditConfig(
            project_root=tmp_path,
            exclude_dirs=set(),
            ignore_params=set(),
            external_filter=True,
            strictness="high",
        ),
    )
    assert isinstance(groups, dict)


def test_analyze_file_local_resolution_globals_only(tmp_path: Path) -> None:
    da = _load()
    path = tmp_path / "mod.py"
    path.write_text(
        textwrap.dedent(
            """
            def foo(a):
                return a

            def outer(x):
                def inner(y):
                    return foo(y)
                return inner(x)
            """
        ).strip()
        + "\n"
    )
    groups, _ = da.analyze_file(
        path,
        recursive=True,
        config=da.AuditConfig(
            project_root=tmp_path,
            exclude_dirs=set(),
            ignore_params=set(),
            external_filter=True,
            strictness="high",
        ),
    )
    assert isinstance(groups, dict)


def test_build_refactor_plan_skips_nontransparent_dependency(tmp_path: Path) -> None:
    da = _load()
    target = tmp_path / "mod.py"
    target.write_text(
        textwrap.dedent(
            """
            def opaque(fn):
                return fn

            @opaque
            def hidden(a, b):
                return a

            def caller(a, b):
                hidden(a, b)
            """
        ).strip()
        + "\n"
    )
    groups_by_path = {target: {"caller": [set(["a", "b"])]}}
    config = da.AuditConfig(
        project_root=tmp_path,
        exclude_dirs=set(),
        ignore_params=set(),
        external_filter=True,
        strictness="high",
        transparent_decorators={"transparent"},
    )
    plan = da.build_refactor_plan(groups_by_path, [target], config=config)
    assert "bundles" in plan



from __future__ import annotations

import argparse
from pathlib import Path
import tomllib


def _read_project_version(pyproject_path: Path) -> str:
    data = tomllib.loads(pyproject_path.read_text(encoding="utf-8"))
    version = data.get("project", {}).get("version")
    if not version:
        raise SystemExit("project.version not found in pyproject.toml")
    return version


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output",
        help="Write key=value output to this path (GitHub Actions output file).",
    )
    args = parser.parse_args()

    version = _read_project_version(Path("pyproject.toml"))
    line = f"version={version}\n"
    if args.output:
        Path(args.output).write_text(line, encoding="utf-8")
    else:
        print(line, end="")


if __name__ == "__main__":
    main()



from __future__ import annotations

import ast
from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from gabion.analysis.dataflow_audit import CallArgs, ParamUse


class ProjectVisitor(ast.NodeVisitor):
    pass


class ParentAnnotator(ast.NodeVisitor):
    def __init__(self) -> None:
        self.parents: dict[ast.AST, ast.AST] = {}

    def generic_visit(self, node: ast.AST) -> None:
        for child in ast.iter_child_nodes(node):
            self.parents[child] = node
            self.visit(child)


class ImportVisitor(ast.NodeVisitor):
    def __init__(self, module_name: str, table) -> None:
        # dataflow-bundle: module_name, table
        self.module = module_name
        self.table = table

    def visit_Import(self, node: ast.Import) -> None:
        for alias in node.names:
            local = alias.asname or alias.name
            self.table.imports[(self.module, local)] = alias.name

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if not node.module and node.level == 0:
            return
        if node.level > 0:
            parts = self.module.split(".")
            if node.level > len(parts):
                return
            base = parts[:-node.level]
            if node.module:
                base.append(node.module)
            source = ".".join(base)
        else:
            source = node.module or ""
        for alias in node.names:
            if alias.name == "*":
                self.table.star_imports.setdefault(self.module, set()).add(source)
                continue
            local = alias.asname or alias.name
            fqn = f"{source}.{alias.name}" if source else alias.name
            self.table.imports[(self.module, local)] = fqn


class UseVisitor(ast.NodeVisitor):
    def __init__(
        self,
        *,
        parents: dict[ast.AST, ast.AST],
        use_map: dict[str, ParamUse],
        call_args: list[CallArgs],
        alias_to_param: dict[str, str],
        is_test: bool,
        strictness: str,
        const_repr: Callable[[ast.AST], str | None],
        callee_name: Callable[[ast.Call], str],
        call_args_factory: Callable[..., CallArgs],
        call_context: Callable[[ast.AST, dict[ast.AST, ast.AST]], tuple[ast.Call | None, bool]],
        return_aliases: dict[str, tuple[list[str], list[str]]] | None = None,
    ) -> None:
        # dataflow-bundle: alias_to_param, call_args, call_args_factory, call_context, callee_name, const_repr, is_test, parents, strictness, use_map
        self.parents = parents
        self.use_map = use_map
        self.call_args = call_args
        self.alias_to_param = alias_to_param
        self.is_test = is_test
        self.strictness = strictness
        self.const_repr = const_repr
        self.callee_name = callee_name
        self.call_args_factory = call_args_factory
        self.call_context = call_context
        self.return_aliases = return_aliases or {}
        self._suspend_non_forward: set[str] = set()
        self._attr_alias_to_param: dict[tuple[str, str], str] = {}
        self._key_alias_to_param: dict[tuple[str, str], str] = {}

    def visit_Call(self, node: ast.Call) -> None:
        callee = self.callee_name(node)
        pos_map: dict[str, str] = {}
        kw_map: dict[str, str] = {}
        const_pos: dict[str, str] = {}
        const_kw: dict[str, str] = {}
        non_const_pos: set[str] = set()
        non_const_kw: set[str] = set()
        star_pos: list[tuple[int, str]] = []
        star_kw: list[str] = []
        for idx, arg in enumerate(node.args):
            if isinstance(arg, ast.Starred):
                if isinstance(arg.value, ast.Name) and arg.value.id in self.alias_to_param:
                    star_pos.append((idx, self.alias_to_param[arg.value.id]))
                else:
                    non_const_pos.add(str(idx))
                continue
            const = self.const_repr(arg)
            if const is not None:
                const_pos[str(idx)] = const
                continue
            if isinstance(arg, ast.Name) and arg.id in self.alias_to_param:
                pos_map[str(idx)] = self.alias_to_param[arg.id]
            else:
                non_const_pos.add(str(idx))
        for kw in node.keywords:
            if kw.arg is None:
                if isinstance(kw.value, ast.Name) and kw.value.id in self.alias_to_param:
                    star_kw.append(self.alias_to_param[kw.value.id])
                else:
                    non_const_kw.add("**")
                continue
            const = self.const_repr(kw.value)
            if const is not None:
                const_kw[kw.arg] = const
                continue
            if isinstance(kw.value, ast.Name) and kw.value.id in self.alias_to_param:
                kw_map[kw.arg] = self.alias_to_param[kw.value.id]
            else:
                non_const_kw.add(kw.arg)
        self.call_args.append(
            self.call_args_factory(
                callee=callee,
                pos_map=pos_map,
                kw_map=kw_map,
                const_pos=const_pos,
                const_kw=const_kw,
                non_const_pos=non_const_pos,
                non_const_kw=non_const_kw,
                star_pos=star_pos,
                star_kw=star_kw,
                is_test=self.is_test,
            )
        )
        self.generic_visit(node)

    def _check_write(self, target: ast.AST) -> None:
        for node in ast.walk(target):
            if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Store):
                name = node.id
                if name in self.alias_to_param:
                    param = self.alias_to_param.pop(name)
                    if param in self.use_map:
                        self.use_map[param].current_aliases.discard(name)
                        self.use_map[param].non_forward = True
                to_remove = [
                    key for key in self._attr_alias_to_param if key[0] == name
                ]
                for key in to_remove:
                    param = self._attr_alias_to_param.pop(key, None)
                    if param in self.use_map:
                        self.use_map[param].non_forward = True
                to_remove = [
                    key for key in self._key_alias_to_param if key[0] == name
                ]
                for key in to_remove:
                    param = self._key_alias_to_param.pop(key, None)
                    if param in self.use_map:
                        self.use_map[param].non_forward = True

    def _bind_sequence(self, target: ast.AST, rhs: ast.AST) -> bool:
        # dataflow-bundle: target, rhs
        if not isinstance(target, (ast.Tuple, ast.List)):
            return False
        if not isinstance(rhs, (ast.Tuple, ast.List)):
            return False
        if len(target.elts) != len(rhs.elts):
            return False
        for lhs, rhs_node in zip(target.elts, rhs.elts):
            if isinstance(lhs, (ast.Tuple, ast.List)) and isinstance(rhs_node, (ast.Tuple, ast.List)):
                if not self._bind_sequence(lhs, rhs_node):
                    self._check_write(lhs)
                continue
            if isinstance(lhs, ast.Name) and isinstance(rhs_node, ast.Name) and rhs_node.id in self.alias_to_param:
                param = self.alias_to_param[rhs_node.id]
                self.alias_to_param[lhs.id] = param
                if param in self.use_map:
                    self.use_map[param].current_aliases.add(lhs.id)
            else:
                self._check_write(lhs)
        return True

    def _collect_alias_sources(self, rhs: ast.AST) -> set[str]:
        if isinstance(rhs, ast.Name) and rhs.id in self.alias_to_param:
            return {self.alias_to_param[rhs.id]}
        if isinstance(rhs, (ast.Tuple, ast.List)):
            sources: set[str] = set()
            for elt in rhs.elts:
                sources.update(self._collect_alias_sources(elt))
            return sources
        return set()

    def _alias_from_call(self, call: ast.Call) -> list[str] | None:
        if not self.return_aliases:
            return None
        callee = self.callee_name(call)
        info = self.return_aliases.get(callee)
        if info is None:
            return None
        params, aliases = info
        if not aliases:
            return None
        mapping: dict[str, str | None] = {}
        for idx, arg in enumerate(call.args):
            if isinstance(arg, ast.Starred):
                return None
            if idx >= len(params):
                return None
            if isinstance(arg, ast.Name) and arg.id in self.alias_to_param:
                mapping[params[idx]] = self.alias_to_param[arg.id]
            else:
                mapping[params[idx]] = None
        for kw in call.keywords:
            if kw.arg is None:
                return None
            if kw.arg not in params:
                continue
            if isinstance(kw.value, ast.Name) and kw.value.id in self.alias_to_param:
                mapping[kw.arg] = self.alias_to_param[kw.value.id]
            else:
                mapping[kw.arg] = None
        resolved: list[str] = []
        for param in aliases:
            mapped = mapping.get(param)
            if not mapped:
                return None
            resolved.append(mapped)
        return resolved

    def _bind_return_alias(
        self, targets: list[ast.AST], aliases: list[str]
    ) -> bool:
        if len(targets) != 1:
            return False
        target = targets[0]
        if isinstance(target, ast.Name):
            if len(aliases) != 1:
                return False
            param = aliases[0]
            self.alias_to_param[target.id] = param
            if param in self.use_map:
                self.use_map[param].current_aliases.add(target.id)
            return True
        if isinstance(target, (ast.Tuple, ast.List)):
            if len(target.elts) != len(aliases):
                return False
            if not all(isinstance(elt, ast.Name) for elt in target.elts):
                return False
            for elt, param in zip(target.elts, aliases):
                if isinstance(elt, ast.Name):
                    self.alias_to_param[elt.id] = param
                    if param in self.use_map:
                        self.use_map[param].current_aliases.add(elt.id)
            return True
        return False

    def visit_Assign(self, node: ast.Assign) -> None:
        if isinstance(node.value, ast.Call):
            aliases = self._alias_from_call(node.value)
            if aliases and self._bind_return_alias(node.targets, aliases):
                self.visit(node.value)
                return
        rhs_param = None
        if isinstance(node.value, ast.Name) and node.value.id in self.alias_to_param:
            rhs_param = self.alias_to_param[node.value.id]

        handled_alias = False
        for target in node.targets:
            if self._bind_sequence(target, node.value):
                handled_alias = True
                continue
            if rhs_param and isinstance(target, ast.Name):
                self.alias_to_param[target.id] = rhs_param
                self.use_map[rhs_param].current_aliases.add(target.id)
                handled_alias = True
            elif rhs_param and isinstance(target, ast.Attribute):
                if isinstance(target.value, ast.Name):
                    self._attr_alias_to_param[(target.value.id, target.attr)] = rhs_param
                    handled_alias = True
            elif rhs_param and isinstance(target, ast.Subscript):
                if (
                    isinstance(target.value, ast.Name)
                    and isinstance(target.slice, ast.Constant)
                    and isinstance(target.slice.value, str)
                ):
                    self._key_alias_to_param[
                        (target.value.id, target.slice.value)
                    ] = rhs_param
                    handled_alias = True
            else:
                self._check_write(target)

        if handled_alias:
            sources = self._collect_alias_sources(node.value)
            self._suspend_non_forward.update(sources)
            self.visit(node.value)
            self._suspend_non_forward.difference_update(sources)
        else:
            self.visit(node.value)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if node.value is None:
            return
        if isinstance(node.value, ast.Call) and isinstance(node.target, ast.Name):
            aliases = self._alias_from_call(node.value)
            if aliases and len(aliases) == 1:
                param = aliases[0]
                self.alias_to_param[node.target.id] = param
                if param in self.use_map:
                    self.use_map[param].current_aliases.add(node.target.id)
                self.visit(node.value)
                return
        rhs_param = None
        if isinstance(node.value, ast.Name) and node.value.id in self.alias_to_param:
            rhs_param = self.alias_to_param[node.value.id]
        handled_alias = False
        if isinstance(node.target, ast.Name) and rhs_param:
            self.alias_to_param[node.target.id] = rhs_param
            self.use_map[rhs_param].current_aliases.add(node.target.id)
            handled_alias = True
        else:
            self._check_write(node.target)
        if handled_alias:
            sources = self._collect_alias_sources(node.value)
            self._suspend_non_forward.update(sources)
            self.visit(node.value)
            self._suspend_non_forward.difference_update(sources)
        else:
            self.visit(node.value)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self._check_write(node.target)
        self.visit(node.value)

    def visit_Name(self, node: ast.Name) -> None:
        if not isinstance(node.ctx, ast.Load):
            return
        if node.id not in self.alias_to_param:
            return
        parent = self.parents.get(node)
        if isinstance(parent, ast.Starred):
            param_name = self.alias_to_param[node.id]
            if self.strictness == "high":
                self.use_map[param_name].non_forward = True
                return
            self.use_map[param_name].direct_forward.add(("args[*]", "arg[*]"))
            return
        if isinstance(parent, ast.keyword) and parent.arg is None:
            param_name = self.alias_to_param[node.id]
            if self.strictness == "high":
                self.use_map[param_name].non_forward = True
                return
            self.use_map[param_name].direct_forward.add(("kwargs[*]", "kw[*]"))
            return
        param_name = self.alias_to_param[node.id]
        if param_name in self._suspend_non_forward:
            return
        call, direct = self.call_context(node, self.parents)
        if call is None or not direct:
            self.use_map[param_name].non_forward = True
            return
        callee = self.callee_name(call)
        slot = None
        for idx, arg in enumerate(call.args):
            if arg is node:
                slot = f"arg[{idx}]"
                break
        if slot is None:
            for kw in call.keywords:
                if kw.value is node and kw.arg is not None:
                    slot = f"kw[{kw.arg}]"
                    break
        if slot is None:
            slot = "arg[?]"
        self.use_map[param_name].direct_forward.add((callee, slot))

    def _root_name(self, node: ast.AST) -> str | None:
        current = node
        while isinstance(current, (ast.Attribute, ast.Subscript)):
            current = current.value
        if isinstance(current, ast.Name):
            return current.id
        return None

    def visit_Attribute(self, node: ast.Attribute) -> None:
        if not isinstance(node.ctx, ast.Load):
            return
        if not isinstance(node.value, ast.Name):
            root_name = self._root_name(node)
            if root_name and root_name in self.alias_to_param:
                param_name = self.alias_to_param[root_name]
                if param_name not in self._suspend_non_forward:
                    self.use_map[param_name].non_forward = True
            self.generic_visit(node)
            return
        key = (node.value.id, node.attr)
        if key not in self._attr_alias_to_param:
            if node.value.id in self.alias_to_param:
                param_name = self.alias_to_param[node.value.id]
                if param_name not in self._suspend_non_forward:
                    self.use_map[param_name].non_forward = True
            return
        param_name = self._attr_alias_to_param[key]
        if param_name in self._suspend_non_forward:
            return
        call, direct = self.call_context(node, self.parents)
        if call is None or not direct:
            self.use_map[param_name].non_forward = True
            return
        callee = self.callee_name(call)
        slot = None
        for idx, arg in enumerate(call.args):
            if arg is node:
                slot = f"arg[{idx}]"
                break
        if slot is None:
            for kw in call.keywords:
                if kw.value is node and kw.arg is not None:
                    slot = f"kw[{kw.arg}]"
                    break
        if slot is None:
            slot = "arg[?]"
        self.use_map[param_name].direct_forward.add((callee, slot))

    def visit_Subscript(self, node: ast.Subscript) -> None:
        if not isinstance(node.ctx, ast.Load):
            return
        if not isinstance(node.value, ast.Name):
            root_name = self._root_name(node)
            if root_name and root_name in self.alias_to_param:
                param_name = self.alias_to_param[root_name]
                if param_name not in self._suspend_non_forward:
                    self.use_map[param_name].non_forward = True
            self.generic_visit(node)
            return
        key_value = None
        if isinstance(node.slice, ast.Constant) and isinstance(node.slice.value, str):
            key_value = node.slice.value
        if key_value is None:
            if node.value.id in self.alias_to_param:
                param_name = self.alias_to_param[node.value.id]
                if param_name not in self._suspend_non_forward:
                    self.use_map[param_name].non_forward = True
            self.visit(node.slice)
            return
        key = (node.value.id, key_value)
        if key not in self._key_alias_to_param:
            if node.value.id in self.alias_to_param:
                param_name = self.alias_to_param[node.value.id]
                if param_name not in self._suspend_non_forward:
                    self.use_map[param_name].non_forward = True
            self.visit(node.slice)
            return
        param_name = self._key_alias_to_param[key]
        if param_name in self._suspend_non_forward:
            return
        call, direct = self.call_context(node, self.parents)
        if call is None or not direct:
            self.use_map[param_name].non_forward = True
            return
        callee = self.callee_name(call)
        slot = None
        for idx, arg in enumerate(call.args):
            if arg is node:
                slot = f"arg[{idx}]"
                break
        if slot is None:
            for kw in call.keywords:
                if kw.value is node and kw.arg is not None:
                    slot = f"kw[{kw.arg}]"
                    break
        if slot is None:
            slot = "arg[?]"
        self.use_map[param_name].direct_forward.add((callee, slot))



from __future__ import annotations

import ast
from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import _analyze_function, _collect_return_aliases
    from gabion.analysis.visitors import ParentAnnotator

    return _analyze_function, _collect_return_aliases, ParentAnnotator


def test_alias_propagation_via_return() -> None:
    _analyze_function, _collect_return_aliases, ParentAnnotator = _load()
    source = textwrap.dedent(
        """
        def ident(x):
            return x

        def sink(v):
            return v

        def foo(a):
            b = ident(a)
            return sink(b)
        """
    )
    tree = ast.parse(source)
    parents = ParentAnnotator()
    parents.visit(tree)
    funcs = [node for node in ast.walk(tree) if isinstance(node, ast.FunctionDef)]
    return_aliases = _collect_return_aliases(funcs, parents.parents, ignore_params=None)
    foo = next(fn for fn in funcs if fn.name == "foo")
    use_map, _ = _analyze_function(
        foo,
        parents.parents,
        is_test=False,
        ignore_params=None,
        strictness="high",
        class_name=None,
        return_aliases=return_aliases,
    )
    assert ("sink", "arg[0]") in use_map["a"].direct_forward



from __future__ import annotations

import ast
from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def test_symbol_table_resolve_star_external_filtered() -> None:
    da = _load()
    table = da.SymbolTable(external_filter=True)
    table.star_imports["pkg.mod"] = {"ext.star"}
    table.module_exports["ext.star"] = {"Foo"}
    table.module_export_map["ext.star"] = {"Foo": "ext.star.Foo"}
    assert table.resolve_star("pkg.mod", "Foo") is None


def test_normalize_callee_long_self_chain() -> None:
    da = _load()
    assert da._normalize_callee("self.a.b", "Service") == "self.a.b"


def test_iter_paths_ignored_file(tmp_path: Path) -> None:
    da = _load()
    path = tmp_path / "skip.py"
    path.write_text("x = 1\n")
    config = da.AuditConfig(exclude_dirs={"skip.py"})
    paths = da._iter_paths([str(path)], config)
    assert paths == []


def test_decorator_name_attribute_without_root() -> None:
    da = _load()
    node = ast.Attribute(value=ast.Constant(value=1), attr="decor")
    assert da._decorator_name(node) is None


def test_normalize_transparent_decorators_variants() -> None:
    da = _load()
    assert da._normalize_transparent_decorators(None) is None
    assert da._normalize_transparent_decorators("a, b") == {"a", "b"}
    assert da._normalize_transparent_decorators(["a", "b, c"]) == {"a", "b", "c"}
    assert da._normalize_transparent_decorators([1, "a"]) == {"a"}
    assert da._normalize_transparent_decorators([]) is None



from __future__ import annotations

import io
import json
from pathlib import Path

import pytest

from gabion.lsp_client import CommandRequest, LspClientError, run_command


def _rpc_message(payload: dict) -> bytes:
    body = json.dumps(payload).encode("utf-8")
    header = f"Content-Length: {len(body)}\r\n\r\n".encode("utf-8")
    return header + body


class _FakeProc:
    def __init__(self, stdout_bytes: bytes, stderr_bytes: bytes, returncode: int | None) -> None:
        self.stdin = io.BytesIO()
        self.stdout = io.BytesIO(stdout_bytes)
        self.stderr = io.BytesIO(stderr_bytes)
        self.returncode = returncode

    def communicate(self, timeout: float | None = None):
        return (b"", self.stderr.read())


def _make_proc(returncode: int | None, stderr_bytes: bytes) -> _FakeProc:
    init = _rpc_message({"jsonrpc": "2.0", "id": 1, "result": {}})
    cmd = _rpc_message({"jsonrpc": "2.0", "id": 2, "result": {}})
    shutdown = _rpc_message({"jsonrpc": "2.0", "id": 3, "result": {}})
    return _FakeProc(init + cmd + shutdown, stderr_bytes, returncode)


def test_run_command_raises_on_nonzero_returncode() -> None:
    def factory(*_args, **_kwargs):
        return _make_proc(1, b"boom")

    with pytest.raises(LspClientError) as exc:
        run_command(CommandRequest("gabion.dataflowAudit", []), root=Path("."), process_factory=factory)
    assert "server failed" in str(exc.value).lower()


def test_run_command_raises_on_stderr_output() -> None:
    def factory(*_args, **_kwargs):
        return _make_proc(0, b"warning")

    with pytest.raises(LspClientError) as exc:
        run_command(CommandRequest("gabion.dataflowAudit", []), root=Path("."), process_factory=factory)
    assert "error output" in str(exc.value).lower()


def test_run_command_allows_blank_stderr() -> None:
    def factory(*_args, **_kwargs):
        return _make_proc(0, b"\n")

    result = run_command(CommandRequest("gabion.dataflowAudit", []), root=Path("."), process_factory=factory)
    assert result == {}



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional, Tuple

Position = Tuple[int, int]


@dataclass(frozen=True)
class TextEdit:
    path: str
    start: Position
    end: Position
    replacement: str


@dataclass(frozen=True)
class FieldSpec:
    name: str
    type_hint: Optional[str] = None


@dataclass(frozen=True)
class RefactorRequest:
    protocol_name: str
    bundle: List[str]
    target_path: str
    fields: List[FieldSpec] = field(default_factory=list)
    target_functions: List[str] = field(default_factory=list)
    compatibility_shim: bool = False
    rationale: Optional[str] = None


@dataclass
class RefactorPlan:
    edits: List[TextEdit] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)
    errors: List[str] = field(default_factory=list)



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_paths, build_synthesis_plan, AuditConfig

    return analyze_paths, build_synthesis_plan, AuditConfig


def test_synthesis_plan_includes_type_hints(tmp_path: Path) -> None:
    analyze_paths, build_synthesis_plan, AuditConfig = _load()
    source = textwrap.dedent(
        """
        def sink(x: int, y: int):
            return x + y

        def g(a, b):
            sink(x=a, y=a)
            sink(x=b, y=b)

        def f(a, b):
            return g(a, b)
        """
    ).lstrip()
    file_path = tmp_path / "mod.py"
    file_path.write_text(source)
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    plan = build_synthesis_plan(
        analysis.groups_by_path,
        project_root=tmp_path,
        max_tier=3,
        min_bundle_size=2,
        allow_singletons=False,
        config=config,
    )
    protocols = plan.get("protocols", [])
    assert protocols
    fields = {field["name"]: field.get("type_hint") for field in protocols[0]["fields"]}
    assert fields.get("a") == "int"
    assert fields.get("b") == "int"


def test_synthesis_plan_resolves_conflicts_to_union(tmp_path: Path) -> None:
    analyze_paths, build_synthesis_plan, AuditConfig = _load()
    source = textwrap.dedent(
        """
        def sink(x: int, y: int):
            return x + y

        def g(a: int, b: int):
            sink(x=a, y=a)
            sink(x=b, y=b)

        def f(a: str, b: str):
            return g(a, b)
        """
    ).lstrip()
    file_path = tmp_path / "mod.py"
    file_path.write_text(source)
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    plan = build_synthesis_plan(
        analysis.groups_by_path,
        project_root=tmp_path,
        max_tier=3,
        min_bundle_size=2,
        allow_singletons=False,
        config=config,
    )
    protocols = plan.get("protocols", [])
    assert protocols
    fields = {field["name"]: field.get("type_hint") for field in protocols[0]["fields"]}
    assert fields.get("a") == "Union[int, str]"
    assert fields.get("b") == "Union[int, str]"


def test_synthesis_plan_infers_types_from_constants(tmp_path: Path) -> None:
    analyze_paths, build_synthesis_plan, AuditConfig = _load()
    source = textwrap.dedent(
        """
        def sink(x=None, y=None):
            return x, y

        def g(a, b):
            sink(x=a, y=a)
            sink(x=b, y=b)

        def f():
            return g(1, "hi")
        """
    ).lstrip()
    file_path = tmp_path / "mod.py"
    file_path.write_text(source)
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    plan = build_synthesis_plan(
        analysis.groups_by_path,
        project_root=tmp_path,
        max_tier=3,
        min_bundle_size=2,
        allow_singletons=False,
        config=config,
    )
    protocols = plan.get("protocols", [])
    assert protocols
    fields = {field["name"]: field.get("type_hint") for field in protocols[0]["fields"]}
    assert fields.get("a") == "int"
    assert fields.get("b") == "str"



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _write(tmp_path: Path, rel: str, content: str) -> Path:
    path = tmp_path / rel
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content))
    return path


def _load_analyzer():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_unused_arg_flow_repo

    return analyze_unused_arg_flow_repo


def test_unused_arg_flow_maps_star_args_low_strictness(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "sample.py",
        """
        def callee(a, b):
            return a

        def caller(*args):
            return callee(*args)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="low",
        external_filter=True,
    )
    assert any("non-constant arg at position 1" in smell for smell in smells)



from __future__ import annotations

import argparse
import subprocess


def _run(cmd: list[str]) -> None:
    subprocess.run(cmd, check=True)


def _git_output(*args: str) -> str:
    result = subprocess.run(["git", *args], check=True, stdout=subprocess.PIPE, text=True)
    return result.stdout.strip()


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", required=True)
    args = parser.parse_args()
    tag = args.tag

    if not tag.startswith("test-v"):
        raise SystemExit(f"Unexpected tag format: {tag}")

    _run(["git", "fetch", "origin", "main", "next", "--tags"])

    tag_sha = _git_output("rev-parse", f"refs/tags/{tag}^{{commit}}")
    main_sha = _git_output("rev-parse", "origin/main")
    next_sha = _git_output("rev-parse", "origin/next")

    if tag_sha != main_sha:
        raise SystemExit(f"Test tag must point at main.\ntag={tag_sha}\nmain={main_sha}")
    if tag_sha != next_sha:
        raise SystemExit(f"Test tag must point at next.\ntag={tag_sha}\nnext={next_sha}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_paths, build_refactor_plan
    from gabion.analysis.dataflow_audit import AuditConfig

    return analyze_paths, build_refactor_plan, AuditConfig


def test_refactor_plan_orders_callee_first(tmp_path: Path) -> None:
    analyze_paths, build_refactor_plan, AuditConfig = _load()
    source = textwrap.dedent(
        """
        def sink(x=None, y=None):
            return x, y

        def g(a, b):
            sink(x=a, y=a)
            sink(x=b, y=b)

        def f(a, b):
            return g(a, b)
        """
    )
    file_path = tmp_path / "mod.py"
    file_path.write_text(source.strip() + "\n")
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    plan = build_refactor_plan(analysis.groups_by_path, [file_path], config=config)
    bundles = plan.get("bundles", [])
    assert bundles
    order = bundles[0].get("order", [])
    assert order == ["mod.g", "mod.f"]



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _write(tmp_path: Path, rel: str, content: str) -> Path:
    path = tmp_path / rel
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content).lstrip())
    return path


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_paths, AuditConfig

    return analyze_paths, AuditConfig


def test_local_definition_overrides_import(tmp_path: Path) -> None:
    analyze_paths, AuditConfig = _load()
    _write(
        tmp_path,
        "other.py",
        """
        def sink(x=None, y=None):
            return x, y
        """,
    )
    a_path = _write(
        tmp_path,
        "a.py",
        """
        from other import sink

        def target(x=None, y=None):
            return x, y

        def sink(p, q):
            target(x=p, y=p)
            target(x=q, y=q)

        def f(a, b):
            return sink(a, b)
        """,
    )
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [a_path, tmp_path / "other.py"],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    groups = analysis.groups_by_path[a_path]
    assert "sink" in groups
    assert "f" in groups
    assert {"p", "q"} in groups["sink"]
    assert {"a", "b"} in groups["f"]



from __future__ import annotations

import ast
from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import CallArgs, ParamUse, _call_context, _callee_name, _const_repr
    from gabion.analysis.visitors import ParentAnnotator, UseVisitor

    return CallArgs, ParamUse, _call_context, _callee_name, _const_repr, ParentAnnotator, UseVisitor


def _call_args_factory(**kwargs):
    CallArgs, *_ = _load()
    return CallArgs(**kwargs)


def _make_visitor(
    tree: ast.AST,
    strictness: str,
    *,
    return_aliases: dict[str, tuple[list[str], list[str]]] | None = None,
):
    CallArgs, ParamUse, _call_context, _callee_name, _const_repr, ParentAnnotator, UseVisitor = _load()
    annotator = ParentAnnotator()
    annotator.visit(tree)
    use_map = {
        "a": ParamUse(direct_forward=set(), non_forward=False, current_aliases={"a"}),
        "b": ParamUse(direct_forward=set(), non_forward=False, current_aliases={"b"}),
        "args": ParamUse(direct_forward=set(), non_forward=False, current_aliases={"args"}),
        "kwargs": ParamUse(direct_forward=set(), non_forward=False, current_aliases={"kwargs"}),
    }
    alias_to_param = {name: name for name in use_map}
    call_args: list[CallArgs] = []
    return_aliases = return_aliases or {
        "ret": (["a", "b"], ["a", "b"]),
        "ret1": (["a"], ["a"]),
    }
    visitor = UseVisitor(
        parents=annotator.parents,
        use_map=use_map,
        call_args=call_args,
        alias_to_param=alias_to_param,
        is_test=False,
        strictness=strictness,
        const_repr=_const_repr,
        callee_name=_callee_name,
        call_args_factory=_call_args_factory,
        call_context=_call_context,
        return_aliases=return_aliases,
    )
    return visitor, use_map, alias_to_param, call_args


def test_usevisitor_star_forwarding_low_strictness() -> None:
    tree = ast.parse(
        "def f(a, b, *args, **kwargs):\n"
        "    g(a, b, *args, **kwargs, k=kwargs)\n"
    )
    visitor, use_map, _, call_args = _make_visitor(tree, strictness="low")
    visitor.visit(tree)
    assert call_args
    assert ("args[*]", "arg[*]") in use_map["args"].direct_forward
    assert ("kwargs[*]", "kw[*]") in use_map["kwargs"].direct_forward


def test_usevisitor_alias_binding_and_non_forward() -> None:
    tree = ast.parse(
        "def ret(a, b):\n"
        "    return a, b\n"
        "def ret1(a):\n"
        "    return a\n"
        "def f(a, b, *args, **kwargs):\n"
        "    x, y = ret(a, b)\n"
        "    z: int = ret1(a)\n"
        "    (u, v) = (a, b)\n"
        "    obj = a\n"
        "    obj.attr = a\n"
        "    record = a\n"
        "    record['key'] = b\n"
        "    val = obj.attr\n"
        "    val2 = record['key']\n"
        "    nested = (a.b).c\n"
        "    a = 1\n"
        "    a += 2\n"
        "    g(*args, **kwargs)\n"
        "    return a\n"
    )
    visitor, use_map, alias_to_param, call_args = _make_visitor(tree, strictness="high")
    visitor.visit(tree)
    assert call_args
    assert alias_to_param.get("x") == "a"
    assert alias_to_param.get("y") == "b"
    assert alias_to_param.get("z") == "a"
    assert alias_to_param.get("u") == "a"
    assert alias_to_param.get("v") == "b"
    assert use_map["a"].non_forward is True
    assert use_map["b"].non_forward is True
    assert use_map["args"].non_forward is True
    assert use_map["kwargs"].non_forward is True


def test_return_alias_binding_tuple_and_rejects_mismatch() -> None:
    tree = ast.parse("def f(a, b):\n    return a, b\n")
    visitor, _, alias_to_param, _ = _make_visitor(tree, strictness="high")
    targets = [ast.Tuple(elts=[ast.Name(id="x", ctx=ast.Store()), ast.Name(id="y", ctx=ast.Store())], ctx=ast.Store())]
    assert visitor._bind_return_alias(targets, ["a", "b"]) is True
    assert alias_to_param["x"] == "a"
    assert alias_to_param["y"] == "b"
    assert visitor._bind_return_alias(targets, ["a"]) is False


def test_alias_from_call_rejects_starred() -> None:
    tree = ast.parse(
        "def ret(a, b):\n"
        "    return a, b\n"
        "def f(a, b, *args):\n"
        "    return ret(*args)\n"
    )
    visitor, _, _, _ = _make_visitor(
        tree,
        strictness="high",
        return_aliases={"ret": (["a", "b"], ["a", "b"])},
    )
    call = next(node for node in ast.walk(tree) if isinstance(node, ast.Call))
    assert visitor._alias_from_call(call) is None


def test_attribute_and_subscript_forwarding() -> None:
    tree = ast.parse(
        "def f(a):\n"
        "    obj.attr = a\n"
        "    record['key'] = a\n"
        "    g(obj.attr)\n"
        "    h(record['key'])\n"
        "    i(record[0])\n"
        "    j(a.b.c)\n"
    )
    visitor, use_map, _, _ = _make_visitor(tree, strictness="high")
    visitor.visit(tree)
    assert ("g", "arg[0]") in use_map["a"].direct_forward
    assert ("h", "arg[0]") in use_map["a"].direct_forward
    assert use_map["a"].non_forward is True


def test_bind_sequence_mismatch_marks_non_forward() -> None:
    tree = ast.parse("def f(a, b):\n    pass\n")
    visitor, use_map, _, _ = _make_visitor(tree, strictness="high")
    target = ast.Tuple(elts=[ast.Name(id="x", ctx=ast.Store())], ctx=ast.Store())
    rhs = ast.Tuple(
        elts=[ast.Name(id="a", ctx=ast.Load()), ast.Name(id="b", ctx=ast.Load())],
        ctx=ast.Load(),
    )
    assert visitor._bind_sequence(target, rhs) is False
    assert use_map["a"].non_forward is False
    assert use_map["b"].non_forward is False


def test_import_visitor_basic_and_relative() -> None:
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import SymbolTable
    from gabion.analysis.visitors import ImportVisitor

    table = SymbolTable()
    visitor = ImportVisitor("pkg.mod", table)
    visitor.visit(ast.parse("import os\nimport pkg.util as util\n"))
    assert table.imports[("pkg.mod", "os")] == "os"
    assert table.imports[("pkg.mod", "util")] == "pkg.util"

    visitor.visit(ast.parse("from .sub import Thing\n"))
    assert table.imports[("pkg.mod", "Thing")] == "pkg.sub.Thing"

    visitor.visit(ast.parse("from .. import Foo\n"))
    assert table.imports[("pkg.mod", "Foo")] == "Foo"

    visitor.visit(ast.parse("from ..star import *\n"))
    assert "star" in table.star_imports.get("pkg.mod", set())

    before = dict(table.imports)
    deep = ImportVisitor("a.b", table)
    deep.visit(ast.parse("from ..... import Nope\n"))
    assert dict(table.imports) == before



from __future__ import annotations

import re
from typing import Iterable

from gabion.synthesis.model import NamingContext


def _camelize(value: str) -> str:
    parts = [p for p in re.split(r"[^a-zA-Z0-9]+", value) if p]
    return "".join(p[:1].upper() + p[1:] for p in parts)


def _normalize_identifier(value: str, fallback: str) -> str:
    cleaned = re.sub(r"[^a-zA-Z0-9_]", "", value)
    if not cleaned:
        return fallback
    if cleaned[0].isdigit():
        return f"{fallback}{cleaned}"
    return cleaned


def suggest_name(fields: Iterable[str], context: NamingContext | None = None) -> str:
    context = context or NamingContext()
    field_list = [f for f in fields if f]
    if not field_list:
        base = context.fallback_prefix
    else:
        frequency = context.frequency
        anchor = max(
            sorted(field_list),
            key=lambda name: (frequency.get(name, 0), len(name)),
        )
        base = _camelize(anchor) or context.fallback_prefix

    base = f"{base}Bundle"
    base = _normalize_identifier(base, context.fallback_prefix)

    name = base
    counter = 2
    existing = set(context.existing_names)
    while name in existing:
        name = f"{base}{counter}"
        counter += 1
    return name
